#include "command.h"

#include "algorithms.h"
#include "clientprot.h"
#include "sockets.h"
#include "settings.h"
#include "botdb.h"


//////// Commands ////////

void gotHelp(Host *h, Player *p, Command *c)
{
	char *tfinal = c->final;

	if (!*c->final)
	{	// No parameter
		if (p->access >= OP_Moderator)
		{
			String s;
			s += "MERVBot ";
			s += getLevelString(p->access);
			s += " commandlist  (Send ::!help <topic> for more info)";
			h->sendPrivate(p, s);
		}

		switch (p->access)
		{
		case OP_Duke:
		case OP_Baron:
		case OP_King:
		case OP_Emperor:
		case OP_RockStar:
		case OP_Q:
		case OP_God:
		case OP_Owner:
			{
				h->sendPrivate(p, "!help owner - Owner commands");
			}
		case OP_SysOp:	/* FALL THRU */
			{
				h->sendPrivate(p, "!help sop  - SysOp commands");
			}
		case OP_SuperModerator:	/* FALL THRU */
			{
				h->sendPrivate(p, "!help smod - SuperModerator commands");
				h->sendPrivate(p, "!help mod  - Moderator commands");
			}
			break;
		case OP_Moderator:
			{
				h->sendPrivate(p, "Basic: !setlogin !where !zone !say !chat !attach !follow !team !spec !ship !turret !awarp");
			}
			break;
		case OP_Limited:
		case OP_Player:
			{
				String s = "Basic: !listop !help";

				// Limited
				if (h->allowLimited)
				{
					s += (p->access == OP_Limited) ? " !give" : " !own";
				}

				// Moderators
				if (h->botInfo.db->findOperator(p->name))
				{
					s += " !login";
				}

				h->sendPrivate(p, s);
			}
		};

		h->imports->talk(makeLocalHelp(p, c));

		return;
	}
	else
	{
		h->botInfo.db->aliasCommand(c->final);
	}

	{	// Display alias(es)
		String alist = h->botInfo.db->getAliasList(c->final);
		if (alist.len)
		{
			String s;
			s += "AKA  ";
			s += c->final;
			s += " ";
			s += alist;
			h->sendPrivate(p, s);
		}

		// Remove ! from c.final
	}

	switch (p->access)
	{
	case OP_Duke:
	case OP_Baron:
	case OP_King:
	case OP_Emperor:
	case OP_RockStar:
	case OP_Q:
	case OP_God:
	case OP_Owner:
		if (c->checkParam("owner") || c->checkParam("all"))
		{
			h->sendPrivate(p, "Basic: !info !autosave !password !closeall !save !read !set !get !alias !killalias !listalias !log");
		}
		else if (c->checkParam("password"))
		{
			h->sendPrivate(p, "!password [bot password] (change or view bot password)");
		}
		else if (c->checkParam("info"))
		{
			h->sendPrivate(p, "!info (describe bot state)");
		}
		else if (c->checkParam("closeall"))
		{
			h->sendPrivate(p, "!closeall (close all bot spawns)");
		}
		else if (c->checkParam("autosave"))
		{
			h->sendPrivate(p, "!autosave <interval> (change time between automatic database backups)");
		}
		else if (c->checkParam("save"))
		{
			h->sendPrivate(p, "!save (manually backup database)");
		}
		else if (c->checkParam("read"))
		{
			h->sendPrivate(p, "!read (manually re-read INI)");
		}
		else if (c->checkParam("set"))
		{
			h->sendPrivate(p, "!set <section>:<key>:<new value> (change INI key)");
		}
		else if (c->checkParam("get"))
		{
			h->sendPrivate(p, "!get <section>:<key> (retrieve INI key)");
		}
		else if (c->checkParam("alias"))
		{
			h->sendPrivate(p, "!alias <command>:<alias> (add an alias)");
		}
		else if (c->checkParam("killalias"))
		{
			h->sendPrivate(p, "!killalias <alias> (remove an alias)");
		}
		else if (c->checkParam("listalias"))
		{
			h->sendPrivate(p, "!listalias [command] (list all command aliases or just those belonging to one command)");
		}
		else if (c->checkParam("log"))
		{
			h->sendPrivate(p, "!log [maximum] (list the last 5 items in the log, or more)");
		}

	case OP_SysOp:	/* FALL THRU */
		if (c->checkParam("sop") || c->checkParam("all"))
		{
			h->sendPrivate(p, "Basic: !close !spawn !load !unload !setbanner !ownbot !error");
		}
		else if (c->checkParam("spawn"))
		{
			h->sendPrivate(p, "!spawn <bot name>  (create a new bot)");
			h->sendPrivate(p, "!spawn -p=password (set name password)");
			h->sendPrivate(p, "!spawn -s=password (set staff *password)");
			h->sendPrivate(p, "!spawn -a=arena    (set destination arena)");
			h->sendPrivate(p, "!spawn -i=DLL      (import functionality from a DLL)");
		}
		else if (c->checkParam("load"))
		{
			h->sendPrivate(p, "!load <DLL filename> (import functionality from a DLL)");
		}
		else if (c->checkParam("unload"))
		{
			h->sendPrivate(p, "!unload (drop DLL)");
		}
		else if (c->checkParam("close"))
		{
			h->sendPrivate(p, "!close [bot name] (close a bot spawn)");
		}
		else if (c->checkParam("ownbot"))
		{
			h->sendPrivate(p, "!ownbot [on/off] (allow or disallow !own and !give system)");
		}
		else if (c->checkParam("error"))
		{
			h->sendPrivate(p, "!error [on/off]   (create an error broadcast channel for mods)");
			h->sendPrivate(p, "!error -c=channel (set error channel name, default is \"error\")");
			h->sendPrivate(p, "!error -g         (generate a server warning message)");
		}
		else if (c->checkParam("setbanner"))
		{
			h->sendPrivate(p, "!setbanner (make the bot use your banner)");
		}

	case OP_SuperModerator:	/* FALL THRU */
		if (c->checkParam("smod") || c->checkParam("all"))
		{
			h->sendPrivate(p, "Basic: !addop !editop !deleteop !go !limit");
		}
		else if (c->checkParam("limit"))
		{
			h->sendPrivate(p, "!limit <numerical level> (remove bot access below given operator level)");
		}
		else if (c->checkParam("go"))
		{
			h->sendPrivate(p, "!go [#]<arena name/public arena number> (change bot's arena)");
			h->sendPrivate(p, "     # (secret arena)");
		}
		else if (c->checkParam("addop"))
		{
			h->sendPrivate(p, "!addop <name>      (create a new operator)");
			h->sendPrivate(p, "!addop -l=level #  (set operator level ::!help levels)");
			h->sendPrivate(p, "!addop -p=password (set password, no password by default)");
			h->sendPrivate(p, "NOTE: Takes effect when the op sends !login [password]");
		}
		else if (c->checkParam("editop"))
		{
			h->sendPrivate(p, "!editop <name>      (edit existing operator, takes effect during !login)");
			h->sendPrivate(p, "!editop -l=level #  (set operator level ::!help levels)");
			h->sendPrivate(p, "!editop -p=password (set password, no password by default)");
			h->sendPrivate(p, "NOTE: Takes effect when the op sends !login [password]");
		}
		else if (c->checkParam("deleteop"))
		{
			h->sendPrivate(p, "!deleteop <name> (remove operator powers)");
			h->sendPrivate(p, "NOTE: Only takes effect after the ex-op leaves the zone");
		}
		else if (c->checkParam("levels"))
		{
			h->sendPrivate(p, "0 Player               Only !listop");
			h->sendPrivate(p, "1 Player(Limited)      Public, limited bot access");
			h->sendPrivate(p, "2 Moderator(Mod)       Control over basic tools");
			h->sendPrivate(p, "3 SuperModerator(SMod) Manages Moderator accounts");
			h->sendPrivate(p, "4 SysOp(SOp)           Control over advanced tools");
			h->sendPrivate(p, "5 Owner(etc)           No restrictions");
		}

	case OP_Moderator:	/* FALL THRU */
		if (c->checkParam("mod") || c->checkParam("all"))
		{
			h->sendPrivate(p, "Basic: !setlogin !where !zone !say !chat !attach !follow !team !spec !ship !turret !awarp");
		}
		else if (c->checkParam("awarp"))
		{
			h->sendPrivate(p, "!awarp [1/0] (turn on/off bot's antiwarp)");
		}
		else if (c->checkParam("version"))
		{
			h->sendPrivate(p, "!version (display bot version)");
		}
		else if (c->checkParam("logout"))
		{
			h->sendPrivate(p, "!logout (release operator powers)");
		}
		else if (c->checkParam("turret"))
		{
			h->sendPrivate(p, "!turret 0/1 (mode: 0=shoots enemies, 1=fires on team command)");
		}
		else if (c->checkParam("setlogin"))
		{
			h->sendPrivate(p, "!setlogin <new password> (change personal login password)");
		}
		else if (c->checkParam("chat"))
		{
			h->sendPrivate(p, "!chat <chan[,chan[,chan]]> (set bot chat(s))");
		}
		else if (c->checkParam("attach"))
		{
			h->sendPrivate(p, "!attach <player name> (make bot attach to a player)");
		}
		else if (c->checkParam("follow"))
		{
			h->sendPrivate(p, "!follow <player name> (make bot follow a player)");
		}
		else if (c->checkParam("ship"))
		{
			h->sendPrivate(p, "!ship <ship type> (make bot change ships)");
		}
		else if (c->checkParam("spec"))
		{
			h->sendPrivate(p, "!spec (make bot spectate the arena)");
		}
		else if (c->checkParam("team"))
		{
			h->sendPrivate(p, "!team <team number> (make bot change teams)");
			h->sendPrivate(p, "!team <player name> (return team of player)");
		}
		else if (c->checkParam("where"))
		{
			h->sendPrivate(p, "!where <player name> (return last known player coordinates)");
		}
		else if (c->checkParam("say"))
		{
			h->sendPrivate(p, "!say <message> (make the bot say something)");
			h->sendPrivate(p, "!say //Team");
			h->sendPrivate(p, "!say ;X;Channel");
			h->sendPrivate(p, "!say :Name:Remote");
		}
		else if (c->checkParam("zone"))
		{
			h->sendPrivate(p, "!zone <message> (broadcast message zone-wide w/ nametag)");
		}

	case OP_Limited:	/* FALL THRU */
	case OP_Player:
		if (c->checkParam("help"))
		{
			h->sendPrivate(p, "!help [topic] (display available commands and what they do)");
			h->sendPrivate(p, "<> required  [] optional  () comments  -s[=assign] switches");
		}
		else if (c->checkParam("disabled"))
		{
			h->sendPrivate(p, "Forget about it! This command is disabled");
		}
		else if (c->checkParam("listop"))
		{
			if (p->access == OP_Player)
			{
				h->sendPrivate(p, "!listop (list online operators)");
			}
			else
			{
				h->sendPrivate(p, "!listop -o (list online and offline operators)");
				h->sendPrivate(p, "        -o (list only online operators)");
				h->sendPrivate(p, "NOTE: %prefix for online operators");
			}
		}
		else if (c->checkParam("login"))
		{
			h->sendPrivate(p, "!login [password] (grant self operator status)");
		}
	};

	h->imports->talk(makeLocalHelp(p, c));

	c->final = tfinal;
}


void gotCommand(Host *h, Player *p, char *m)
{

#ifdef DISABLE_COMMANDS
	return;											// Compiler directive to ignore commands
#endif

	if (p->access < h->lowestLevel)
	{	// Limiter is in effect
		opEntry *op = h->botInfo.db->findOperator(p->name);

		if (!op || (op->getAccess() < h->lowestLevel))
		{
			return;
		}
	}

	Command c(m);									// Wrap in a friendly class

	char *tcmd = c.cmd;							// Backup original pointer in case this is an alias
	h->botInfo.db->aliasCommand(c.cmd);				// Check aliases

try {												// Catch all command-crash bugs

	switch (p->access)
	{
	case OP_Duke:
	case OP_Baron:
	case OP_King:
	case OP_Emperor:
	case OP_RockStar:
	case OP_Q:
	case OP_God:
	case OP_Owner:
		if (c.check("password"))
		{
			String s;
			if (*c.final)
			{
				s += "?password=";
				s += c.final;
				h->sendPublic(s);

				h->botInfo.setLogin(h->botInfo.name, c.final, h->botInfo.staffpw);

				h->sendPrivate(p, "Updated local bot parameters and requested network password change");
			}
			else
			{
				s += "Current bot password: ";
				s += h->botInfo.password;
				h->sendPrivate(p, s);
			}
		}
		else if (c.check("info"))
		{
			String s;

			s += "MISC SysOp:";
			s += h->hasSysOp;
			s += " SMod:";
			s += h->hasSMod;
			s += " Mod:";
			s += h->hasMod;
			s += " Ping:";
			s += h->syncPing;
			s += " Client:";
			if (h->botInfo.db->forceContinuum)
			{
				s += "Ctm ";
				s += CONTINUUM_VERSION;
			}
			else
			{
				s += "VIE ";
				s += SUBSPACE_VERSION;
			}
			s += " DLL:";
			s += h->botInfo.dllName;
			s += " DELT:";
			s += h->timeDiff;
			h->sendPrivate(p, s);

			s = "USER ";
			s += h->botInfo.name;
			s += ":";
			s += h->botInfo.password;

			s += "  ZONE ";
			{	// Get string IP
				INADDR addr(h->botInfo.ip, 0);
				s += addr.getString();
			}
			s += ":";
			s += h->botInfo.port;

			s += "  ARENA \'";
			s += h->botInfo.initialArena;
			s += "\' ship #";
			s += h->botInfo.initialShip;
			s += " @ ";
			s += h->botInfo.xres;
			s += "x";
			s += h->botInfo.yres;
			h->sendPrivate(p, s);

			s = "BAN  MID:";
			s += h->botInfo.machineID;
			s += " PID:";
			s += h->botInfo.permissionID;
			s += " TZB:";
			s += h->botInfo.timeZoneBias;
			h->sendPrivate(p, s);
		}
		else if (c.check("closeall"))
		{
			h->botInfo.db->spawns.massDisconnect();

			break;
		}
		else if (c.check("log"))
		{
#ifdef KEEP_LOG
			int max = 5;

			if (isNumeric(c.final))
			{
				max = getInteger(c.final, 10);
				if (max < 1) max = 5;
			}

			if (max > h->logLength) max = h->logLength;

			for (int i = h->logLength - max; i < h->logLength; ++i)
			{
				h->sendPrivate(p, *h->loggedChatter[i]);
			}

			break;
#endif
		}
		else if (c.check("autosave"))
		{
			if (isNumeric(c.final))
			{
				Uint32 interval = getInteger(c.final, 10);

				if ((interval >= 30) && (interval <= 10000000))
				{
					h->botInfo.db->saveInterval = interval * 100;
				}

				String s;
				s += "Database autosave every ";
				s += h->botInfo.db->saveInterval;
				s += " seconds";
				h->sendPrivate(p, s);
			}
			else
			{
				h->sendPrivate(p, "Format: !autosave <database interval in seconds>");
			}
		}
		else if (c.check("ppl"))	// unlisted, spit out diag report
		{
			_listnode <Player> *parse = h->playerlist.head;
			String s;
			Uint32 count = 0;

			while (parse)
			{
				Player *p = parse->item;
				parse = parse->next;

				s += p->name;
				s += "(";
				s += p->score.wins;
				s += ":";
				s += p->score.losses;
				s += ")@";
				s += p->team;
				++count;

				if (parse)
				{
					if (s.len > 80)
					{
						h->sendPublic(s);
						s = "";
						count = 0;
					}
					else
					{
						s += "  ";
					}
				}
			}

			if (count) h->sendPublic(s);
		}
		else if (c.check("save"))
		{
			h->botInfo.db->saveOperators();
			h->botInfo.db->saveAliases();

			h->sendPrivate(p, "Database save completed.");
		}
		else if (c.check("read"))
		{
			h->botInfo.db->reloadINI(false);

			h->sendPrivate(p, "Re-loaded INI settings.");
		}
		else if (c.check("set"))
		{
			String out = c.final;

			String app, key, str;

			app = out.split(':');
			key = out.split(':');
			str = out.split(':');

			WritePrivateProfileString(app, key, str, h->botInfo.db->path);

			String s;
			s += "Set->";
			s += app;
			s += ":";
			s += key;
			s += ":";
			s += str;

			h->sendPrivate(p, s);
		}
		else if (c.check("get"))
		{
			String out = c.final;

			String app, key;

			app = out.split(':');
			key = out.split(':');

			char buffer[128];
			GetPrivateProfileString(app, key, "InvalidTag", buffer, 128, h->botInfo.db->path);
			buffer[127] = '\0';

			String s;
			s += "Get->";
			s += app;
			s += ":";
			s += key;
			s += ":";
			s += buffer;

			h->sendPrivate(p, s);
		}
		else if (c.check("alias"))
		{
			tolower(c.final);

			String out = c.final;

			String command, alias;

			command = out.split(':');
			alias = out.split(':');

			h->botInfo.db->aliasCommand(command.msg);

			if (!invalidArena(command) && !invalidArena(alias))
			{
				if (h->botInfo.db->findAlias(alias))
				{
					String s;
					s += "Alias \'";
					s += alias;
					s += "\' already exists";
					h->sendPrivate(p, s);
				}
				else
				{
					h->botInfo.db->addAlias(command, alias);

					String s;
					s += "Created alias \'";
					s += alias;
					s += "\' for command \'";
					s += command;
					s += "\'";
					h->sendPrivate(p, s);
				}
			}
			else
			{
				String s;
				s += "Invalid command or alias \'";
				s += alias;
				s += "\':\'";
				s += command;
				s += "\'";
				h->sendPrivate(p, s);
			}
		}
		else if (c.check("killalias"))
		{
			tolower(c.final);

			String s;
			if (h->botInfo.db->killAlias(c.final))
			{
				s += "Removed alias \'";
				s += c.final;
				s += "\'";
			}
			else
			{
				s += "Alias \'";
				s += c.final;
				s += "\' does not exist";
			}
			h->sendPrivate(p, s);
		}
		else if (c.check("listalias"))
		{
			String s;
			Uint32 count = 0;
			s += " ";
			char *cmd = c.final;

			if (*cmd)
			{
				h->botInfo.db->aliasCommand(cmd);

				s += cmd;
				++count;
			}
			else
			{
				break;
			}

			_listnode <cmdAlias> *parse = h->botInfo.db->aliasList.head;

			while (parse)
			{
				cmdAlias *alias = parse->item;
				parse = parse->next;

				if (!alias->isCmd(cmd)) continue;

				if (count++) s += " ";
				s += alias->getAlias();

				if (s.len > 80)
				{
					h->sendPrivate(p, s);
					s = " ";
					count = 0;
				}
			}

			if (count)
			{
				h->sendPrivate(p, s);
			}
		}

	case OP_SysOp:	/* FALL THRU */
		if (c.check("spawn"))
		{
			if (h->botInfo.db->spawns.getConnections() >= h->botInfo.maxSpawns)
			{
				h->sendPrivate(p, "Aborted: Too many bot spawns are connected or connecting. Try again soon");

				break;
			}

			BOT_INFO bi;
			bi.set(h->botInfo);

			String name;
			String password;
			String staff;
			String arena;
			String dll;

			_switch *s;

			// Name
			if (invalidName(c.final))
			{
				h->sendPrivate(p, "Aborted: The bot name you gave me is invalid");

				break;
			}
			name = c.final;

			// Password
			if (s = c.getParam('p'))
			{
				password = s->param;
			}
			else
			{
				password = bi.password;
			}

			// Staff
			if (s = c.getParam('s'))
			{
				staff = s->param;
			}
			else
			{
				staff = bi.staffpw;
			}

			// Imports
			if (s = c.getParam('i'))
			{
				dll = s->param;
			}
			else
			{
				dll = bi.dllName;
			}

			// Arena
			if (s = c.getParam('a'))
			{
				arena = s->param;
			}
			else
			{
				arena = bi.initialArena;
			}

			if (invalidArena(arena))
			{
				h->sendPrivate(p, "Aborted: Arena name you gave me is invalid");

				break;
			}

			bi.setLogin(name, password, staff);
			bi.setSpawn(dll);
			bi.setArena(arena, bi.initialShip, bi.xres, bi.yres, bi.allowAudio);

			String m;

			if (bi.db->spawns.connectHost(bi))
			{
				m += "Attempting to spawn ";
				m += name;
				m += ". If successful, it will be in arena \'";
				m += arena;
				m += "\' in a few seconds...";
			}
			else
			{
				m += "Error while attempting to spawn ";
				m += name;
				m += ". A bot with this name is already active";
			}
			h->sendPrivate(p, m);
		}
		else if (c.check("load"))
		{
			if (!*c.final)
			{
				h->sendPrivate(p, "Syntax: !load <filename>");
				break;
			}

			h->imports->clearImports();				// Release callbacks

			if (h->imports->importLibrary(c.final))	// Attempt to load new callbacks
			{
				String s;
				s += "Successfully loaded module \'";
				s += c.final;
				s += "\'.";
				h->sendPrivate(p, s);

				h->botInfo.setSpawn(c.final);
			}
			else
			{
				h->sendPrivate(p, "Old plugin unloaded. I was unable to find the plugin you requested");
			}
		}
		else if (c.check("unload"))
		{
			h->imports->clearImports();				// Release callbacks

			h->sendPrivate(p, "Old plugin unloaded");
		}
		else if (c.check("setbanner"))
		{
			h->postRR(generateChangeBanner(p->banner));

			h->sendPublic("*banner please");
		}
		else if (c.check("ownbot"))
		{
			if (c.checkParam("on"))
			{
				h->allowLimited = true;
			}
			else if (c.checkParam("off"))
			{
				h->allowLimited = false;
				h->revokeAccess(OP_Limited);
			}

			if (h->allowLimited)
			{
				h->sendPrivate(p, "!own and !give commands allowed, listed in !help");
			}
			else
			{
				h->sendPrivate(p, "!own and !give commands disallowed");
			}
		}
		else if (c.check("close"))
		{
			if (h->botInfo.db->spawns.getConnections() <= 1)
			{
				h->sendPrivate(p, "Aborted: Too few bot spawns are connected or connecting. Try again soon");

				break;
			}

			h->disconnect(true);
		}
		else if (c.check("error"))
		{
			if (c.getParam('g'))
			{
				h->postRR(generateViolation(SEC_ShipChecksumMismatch));

				break;
			}

			String chan;

			_switch *i;
			if (i = c.getParam('c'))
			{
				chan = i->param;
			}
			else
			{
				chan = "error";
			}

			if (c.checkParam("on"))
			{
				h->broadcastingErrors = true;
			}
			else if (c.checkParam("off"))
			{
				h->broadcastingErrors = false;
			}

			String change = "?chat=";
			change += chan;
			h->sendPublic(change);

			String s = "Error broadcasting ";
			if (h->broadcastingErrors)	s += "on, ";
			else						s += "off, ";
			s += "sending to channel \"";
			s += chan;
			s += "\"";

			h->sendPrivate(p, s);
		}

	case OP_SuperModerator:	/* FALL THRU */
		if (c.check("limit"))
		{
			if (*c.final)
			{
				if (!isNumeric(c.final))
				{
					h->sendPrivate(p, "Format: !limit <numerical operator level>");

					break;
				}

				Uint16 limit = getInteger(c.final, 10);
				if (limit > OP_Owner) limit = OP_Owner;

				if (limit > p->access)
				{
					h->sendPrivate(p, "Denied! Commiting bot suicide ain't cool");

					break;
				}

				h->lowestLevel = (Operator_Level)limit;
			}

			String s;
			s += "Limiting bot access to ";
			s += getLevelString(h->lowestLevel);
			s += " or higher";
			h->sendPrivate(p, s);
		}
		else if (c.check("addop"))
		{
			// Only valid names
			if (invalidName(c.final))
			{
				String s;
				s += "Invalid operator name \'";
				s += c.final;
				s += "\'";
				h->sendPrivate(p, s);

				break;
			}

			opEntry *op;
			Operator_Level access;

			// Only new names
			op = h->botInfo.db->findOperator(c.final);

			if (op)
			{
				h->sendPrivate(p, "Cannot recreate an existing operator.  ::!listop  ::!help editop");

				break;
			}

			// Process level switch
			_switch *i;
			if ((i = c.getParam('l')) && isNumeric(i->param))
			{
				access = (Operator_Level)getInteger(i->param, 10);
				if (access < 0)
				{
					access = OP_Moderator;

					break;
				}
				switch (p->access)
				{
				case OP_SuperModerator:
					access = OP_Moderator;
					break;
				case OP_SysOp:
					if (access >= OP_Owner) access = OP_SysOp;
				};
				if (access == OP_Player)	access = OP_Moderator;
				else if (access > OP_God)	access = OP_God;
			}
			else
			{
				access = OP_Moderator;
			}

			// Add operator
			op = h->botInfo.db->addOperator(c.final, "", access);

			if (!op)
			{
				h->sendPrivate(p, "The fates have conspired against you: addOperator() failed");

				break;
			}

			// Notify requesting operator
			String s;
			s += "Created ";
 			s += getLevelString(op->getAccess());
			s += ": ";
			s += op->getName();
			if ((i = c.getParam('p')) && (i->param[0]))
			{
				s += ", with password ";
				s += i->param;

				op->setPassword(i->param);
			}
			else
			{
				s += ", without a password";
			}

			h->sendPrivate(p, s);
		}
		else if (c.check("go"))
		{
			if (invalidArena(c.final))
			{
				String s;
				s += "Denied! Invalid arena name \'";
				s += c.final;
				s += "\'";
				h->sendPrivate(p, s);

				break;
			}

			String s;
			s += "Moving to \'";
			s += c.final;
			s += "\'";
			h->sendRemotePrivate(p->name, s);

			h->botInfo.setArena(c.final, h->botInfo.initialShip, h->botInfo.xres, h->botInfo.yres, h->botInfo.allowAudio);

			h->changeArena(c.final);
		}
		else if (c.check("editop"))
		{
			opEntry *op;

			op = h->botInfo.db->findOperator(c.final);

			if (!op)
			{
				h->sendPrivate(p, "Cannot modify non-existant operators.  ::!listop  ::!help addop");

				break;
			}

			_switch *i;
			if ((i = c.getParam('l')) && isNumeric(i->param))
			{
				Operator_Level access = (Operator_Level)getInteger(i->param, 10);
				if (access < 0)
				{
					access = OP_Moderator;

					break;
				}
				switch (p->access)
				{
				case OP_SuperModerator:
					access = OP_Moderator;
					break;
				case OP_SysOp:
					if (access >= OP_Owner) access = OP_SysOp;
				};
				if (access == OP_Player)	access = OP_Moderator;
				else if (access > OP_God)	access = OP_God;

				op->setAccess(access);
			}

			// Notify requesting operator
			String s;
			s += "Updated. ";
 			s += getLevelString(op->getAccess());
			s += ": ";
			s += op->getName();
			if (i = c.getParam('p'))
			{
				op->setPassword(i->param);

				if (i->param[0])
				{
					s += ", changed password ";
					s += i->param;

					op->setPassword(i->param);
				}
				else
				{
					s += ", removed the password";
				}
			}

			h->sendPrivate(p, s);
		}
		else if (c.check("deleteop"))
		{
			// Determine which operator is being requested removed
			opEntry *xop = h->botInfo.db->findOperator(c.final);

			if (!xop)
			{
				h->sendPrivate(p, "Removing a non-existant operator isn't groovy baby");

				break;
			}

			if (CMPSTR(xop->getName(), p->name))
			{
				h->sendPrivate(p, "Thou shalt not commit bot suicide");

				break;
			}

			// Run the request through level restrictions
			bool failed = false;

			switch (p->access)
			{
			case OP_SuperModerator:
				if (xop->getAccess() != OP_Moderator)
				{
					h->sendPrivate(p, "SuperModerators may only remove Moderators");

					failed = true;
				}
				break;
			case OP_SysOp:
				if (xop->getAccess() > OP_SysOp)
				{
					h->sendPrivate(p, "SysOps may not remove Owner-equivalent operators");

					failed = true;
				}
			};

			if (failed) break;

			// Remove the operator
			if (h->botInfo.db->removeOperator(xop->getName()))
			{
				h->sendPrivate(p, "Successfully removed operator");
			}
			else
			{
				h->sendPrivate(p, "Failed to remove operator");
			}
		}

	case OP_Moderator:	/* FALL THRU */
		if (c.check("setlogin"))
		{
			opEntry *op = h->botInfo.db->findOperator(p->name);
			op->setPassword(c.final);

			String s;

			if (c.checkParam(""))
			{
				s += "Login password no longer required";
			}
			else
			{
				s += "Login password changed to ";
				s += c.final;
			}

			h->sendPrivate(p, s);
		}
		else if (c.check("version"))	// unlisted, display bot version
		{
			h->sendPrivate(p, VERSION_STRING);
		}
		else if (c.check("awarp"))
		{
			if (h->Me)
			{
				if (c.checkParam("0"))
					h->Me->awarp = 0;
				else if (c.checkParam("1"))
					h->Me->awarp = 1;

				switch (h->Me->awarp)
				{
				case 0:		h->sendPrivate(p, "Mode set: 0, I will not engage antiwarp.");			break;
				case 1:		h->sendPrivate(p, "Mode set: 1, I will project an anti-warp field around my ship.");	break;
				};
			}
		}
		else if (c.check("turret"))
		{
			if (c.checkParam("0"))
				h->turretMode = 0;
			else if (c.checkParam("1"))
				h->turretMode = 1;

			switch (h->turretMode)
			{
			case 0:		h->sendPrivate(p, "Mode set: 0, when in-game and not following I will fire against enemy teams");			break;
			case 1:		h->sendPrivate(p, "Mode set: 1, when in-game and not following I will act like a team-controlled turret");	break;
			default:	h->sendPrivate(p, "Unknown mode set.  This be a problem you should talk to Catid about");					break;
			};
		}
		else if (c.check("drop"))	// unlisted, drops carried flags
		{
			h->postRR(generateFlagDrop());
		}
		else if (c.check("where"))
		{
			Player *xp;

			if (!*c.final)
				xp = p;
			else
				xp = h->findPlayer(c.final);

			if (xp)
			{
				Sint32 x = xp->tile.x,
					   y = xp->tile.y;

				String s;
				s += "Last seen ";
				s += xp->name;
				s += " @ (";
				s += x;
				s += ", ";
				s += y;
				s += ") [";
				s += getCoords(x, y);
				s += "]";
				h->sendPrivate(p, s);
			}
		}
		else if (c.check("spec"))
		{
			h->postRR(generateChangeShip(SHIP_Spectator));

			h->botInfo.setArena(h->botInfo.initialArena, SHIP_Spectator, h->botInfo.xres, h->botInfo.yres, h->botInfo.allowAudio);
		}
		else if (c.check("follow"))
		{
			Player *xp;
			String s;

			if (!*c.final)
			{
				if (h->follow)
				{
					s += "Currently following ";
					s += h->follow->item->name;
				}
				else
				{
					s += "Currently not following anyone";
				}

				s += ". Type !follow <name> to start following";

				h->sendPrivate(p, s);

				break;
			}

			if (c.checkParam("on"))
			{	// because people ignore the !help reference
				xp = p;
			}
			else
			{
				xp = h->findPlayer(c.final);
			}

			if (xp)
			{
				h->follow = h->playerlist.find(xp);

				s += "Started following ";
				s += xp->name;
				s += ". Will start once player is visible";

				if (h->Me)
				{
					h->Me->clone(xp);

					if (h->Me->ship == SHIP_Spectator)
					{
						h->postRR(generateChangeShip(SHIP_Warbird));

						h->botInfo.setArena(h->botInfo.initialArena, SHIP_Warbird, h->botInfo.xres, h->botInfo.yres, h->botInfo.allowAudio);
					}
				}

				h->sendPrivate(p, s);
			}
			else
			{
				if (h->follow)
				{
					s += "Stopped following ";
					s += h->follow->item->name;

					if (h->Me)
					{
						h->Me->vel = 0;
					}

					h->follow = NULL;
				}
				else
				{
					s += "Stopped following";
				}

				h->sendPrivate(p, s);
			}
		}
		else if (c.check("attach"))
		{
			Player *xp;

			if (!*c.final || c.checkParam("on"))
			{	// because people ignore the !help reference
				xp = p;
			}
			else
			{
				xp = h->findPlayer(c.final);
			}

			if (xp)
			{
				String s;
				s += "Attaching to ";
				s += xp->name;
				s += "...";
				h->sendPrivate(p, s);

				h->postRR(generateAttachRequest(xp->ident));

				h->follow = h->playerlist.find(xp);
			}
			else
			{
				h->sendPrivate(p, "Detached");

				h->postRR(generateAttachRequest(UNASSIGNED));
			}
		}
		else if (c.check("ship"))
		{
			Uint8 ship = (getInteger(c.final, 10) - 1) & 7;

			h->botInfo.setArena(h->botInfo.initialArena, (Ship_Types)ship, h->botInfo.xres, h->botInfo.yres, h->botInfo.allowAudio);

			h->postRR(generateChangeShip((Ship_Types)ship));
		}
		else if (c.check("team"))
		{
			if (isNumeric(c.final))
			{
				Uint16 team = getInteger(c.final, 10);
				team %= MAX_TEAMS;

				h->postRR(generateChangeTeam(team));
			}
			else
			{
				String s;
				Player *px;

				if (!*c.final)
				{
					px = p;
				}
				else
				{
					px = h->findPlayer(c.final);
				}

				if (px)
				{
					s += "Found ";
					s += px->name;
					s += " on team ";
					s += px->team;
				}
				else
				{
					s += "Unable to find ";
					s += c.final;
				}

				h->sendPrivate(p, s);
			}
		}
		else if (c.check("chat"))
		{
			if (*c.final)
			{
				if (h->broadcastingErrors)
				{
					h->sendPrivate(p, "Cannot change chat channel while broadcasting errors");
				}
				else
				{
					String s = "?chat=";
					s += c.final;
					h->sendPublic(s);

					s = "chat=";
					s += c.final;
					h->sendPrivate(p, s);
				}
			}
			else
			{
				h->sendPrivate(p, "You must provide channel name(s): !chat squad_chat,newbie_chat,zone_chat");
			}
		}
		else if (c.check("say"))
		{
			switch (c.final[0])
			{
			case '/':	// Team
				if (c.final[1] != '/') break;

				if ( (c.final[2] != '*') &&
					 (c.final[2] != '?')  )
				{
					h->sendTeam(c.final + 2);
				}
				break;
			case '\'':	// Team
				if ( (c.final[1] != '*') &&
					 (c.final[1] != '?')  )
				{
					h->sendTeam(c.final + 1);
				}
				break;
			case ';':	// Channel
				if ( (c.final[1] != '*') &&
					 (c.final[1] != '?')  )
				{
					h->sendChannel(c.final + 1);
				}
				break;
			case ':':	// Remote private
				if (validRemotePrivate(c.final))
				{
					h->sendRemotePrivate(c.final);
				}
				break;
			default:	// Public
				if ( (c.final[0] != '*') &&
					 (c.final[0] != '?')  )
				{
					h->sendPublic(c.final);
				}
			};
		}
		else if (c.check("zone"))
		{
			if (!*c.final)
			{
				h->sendPrivate(p, "Here's how \'!zone\' works: !zone <message>  (I will add a -nametag to the end)");

				break;
			}

			String s;
			s += "*zone";
			s += c.final;
			s += " -";
			s += p->name;
			h->sendPublic(s);
		}
		else if (c.check("kill"))	// disabled in commands.txt by default
		{
			h->postRR(generateDeath(p->ident, 500));
		}
		else if (c.check("logout"))	// unlisted, remove operator powers
		{
			p->access = OP_Player;
		}

	case OP_Limited:	/* FALL THRU */
		if (h->allowLimited && c.check("give"))
		{
			h->revokeAccess(OP_Limited);
			h->limitedOwner = NULL;

			Player *xp = h->findPlayer(c.final);
			if (xp && (xp->access == OP_Player))
			{
				xp->access = OP_Limited;

				String s = "You have been granted bot ownership powers by ";
				s += p->name;
				s += ". Send ::!help for extended commands";
				h->sendPrivate(xp, s);

				h->sendPrivate(p, "Your ownership powers have been successfully transferred");
			}
			else
			{
				h->sendPrivate(p, "Your ownership powers have been successfully released");
			}
		}

	case OP_Player:	/* FALL THRU */
		if (c.check("help"))
		{
			gotHelp(h, p, &c);
		}
		else if (c.check("disabled"))
		{
			h->sendPrivate(p, "Forget about it! This command is disabled");
		}
		else if (c.check("lag"))	// unlisted, display session latency
		{
			// PING Current:BLEH ms  Average:BLEH ms  Low:BLEH ms  High:BLEH ms
			String s;
			s += "PING Current:";
			s += h->syncPing;
			s += " ms  Average:";
			s += h->avgPing;
			s += " ms  Low:";
			s += h->lowPing;
			s += " ms  High:";
			s += h->highPing;
			s += " ms  Delta:";
			s += h->timeDiff;
			s += " ms";

			h->sendPrivate(p, s);
		}
		else if (c.check("sex"))	// unlisted, have sex with bot
		{
			h->sendPrivate(p, SND_Girl, "Unf! Unf! Unf!");
		}
		else if (c.check("sheep"))	// unlisted, sheeply bah
		{
			h->sendPrivate(p, SND_Sheep, "Bah!");
		}
		else if (h->allowLimited && c.check("own"))
		{
			Player *xp = h->limitedOwner;
			if (xp)
			{
				String s;
				s += "Player ";
				s += xp->name;
				s += " has claimed me";

				h->sendPrivate(p, s);
			}
			else if (p->access == OP_Player)
			{
				h->sendPrivate(p, "You are now my owner. Send ::!help for extended commands");

				p->access = OP_Limited;
				h->limitedOwner = p;
			}
			else
			{
				h->sendPrivate(p, "You cannot claim Limited ownership as an operator of higher rank");
			}
		}
		else if (c.check("listop"))
		{
			Operator_Level access = OP_Player;
			Uint32 total = 0;
			bool onlineOnly;
			String l;

			// compute onlineOnly
			if (p->access == OP_Player)
			{
				onlineOnly = true;
			}
			else
			{
				onlineOnly = (c.getParam('o') != NULL);
			}

			if (onlineOnly)
			{	// only online operators
				_listnode <Player> *parse = h->playerlist.head;

				while (parse)
				{
					Player *o = parse->item;
					parse = parse->next;

					if (o->access == OP_Player) continue;

					if (access != o->access)
					{
						l += getLevelString(access = o->access);
						l += ": ";
					}

					l += o->name;

					++total;

					if (parse)
					{
						if (l.len > 80)
						{
							h->sendPrivate(p, l);
							l.clear();
							access = OP_Player;
						}
						else
						{
							l += "  ";
						}
					}
				}
			}
			else
			{	// online and offline operators
				_listnode <opEntry> *parse = h->botInfo.db->opList.head;

				while (parse)
				{
					opEntry *o = parse->item;
					parse = parse->next;

					if (access != o->getAccess())
					{
						l += getLevelString(access = o->getAccess());
						l += ": ";
					}

					Player *xp = h->findPlayer(o->getName());
					if (xp && (xp->access != OP_Player))
						l += "%";

					l += o->getName();

					++total;

					if (parse)
					{
						if (l.len > 80)
						{
							h->sendPrivate(p, l);
							l.clear();
							access = OP_Player;
						}
						else
						{
							l += "  ";
						}
					}
				}
			}

			if (total)
				h->sendPrivate(p, l);
			else
				h->sendPrivate(p, "No operators");
		}
		else if (c.check("login"))
		{
			opEntry *op = h->botInfo.db->findOperator(p->name);

			if (op)
			{	// Listed
				if (!h->billerOnline && op->validatePass(""))
				{
					op->addFailure();

					h->sendPrivate(p, "Unable to log you on while the biller is down. This failure has been logged");

					break;
				}

				if (op->validatePass(c.final))
				{	// Grant access

					if (p == h->limitedOwner) h->limitedOwner = NULL;	// reset Limited ownership

					p->access = op->getAccess();
					op->addCounter();

					Uint32 failure = op->getFailureCount();
					Uint32 overall = op->getOverallCount();

					String s;
					s += getLevelString(p->access);
					s += ": Welcome. Sessions  ";

					s += overall;
					if (overall == 1)	s += " logged  ";
					else				s += " logged  ";

					s += failure;
					if (failure == 1)	s += " failed ";
					else				s += " failed ";

					s += " from your handle";

					h->sendPrivate(p, s);
				}
				else
				{	// Bad password
					op->addFailure();

					h->sendPrivate(p, "Bad password. This failure has been logged");

					break;
				}
			}
			else if (p->access > OP_Player)
			{	// Not listed, but has powers already
				h->sendPrivate(p, "Login failure: account revoked. Please contact your senior staff");
			}
		}
	};

	h->imports->talk(makeLocalCommand(p, &c));

} catch (...)	// Catch command-crash bugs
{
	h->sendPrivate(p, SND_Inconceivable, "Oh no!  You've managed to find a command-crash bug!  Report to Catid@pacbell.net");
}

	c.cmd = tcmd;	// Restore pointer from alias string
}

void gotRemoteHelp(Host *h, char *p, Command *c, Operator_Level l)
{
	char *tfinal = c->final;						// Backup original pointer in case this is an alias

	if (!*c->final)
	{
		switch (l)
		{
		case OP_Duke:
		case OP_Baron:
		case OP_King:
		case OP_Emperor:
		case OP_RockStar:
		case OP_Q:
		case OP_God:
		case OP_Owner:
		case OP_SysOp:
			h->sendRemotePrivate(p, "Commands: !help !listop !chat !version !go !close");
			break;
		case OP_SuperModerator:	/* FALL THRU */
			h->sendRemotePrivate(p, "Commands: !help !listop !chat !version !go");
			break;
		case OP_Moderator:	/* FALL THRU */
			h->sendRemotePrivate(p, "Commands: !help !listop !chat !version");
			break;
		};

		h->imports->talk(makeRemoteHelp(p, c, l));

		return;
	}
	else
	{
		h->botInfo.db->aliasCommand(c->final);			// Check aliases
	}

	switch (l)
	{
	case OP_Duke:
	case OP_Baron:
	case OP_King:
	case OP_Emperor:
	case OP_RockStar:
	case OP_Q:
	case OP_God:
	case OP_Owner:
	case OP_SysOp:
		if (c->checkParam("close"))
		{
			h->sendRemotePrivate(p, "!close (close all but last bot spawn)");
		}

	case OP_SuperModerator:	/* FALL THRU */
		if (c->checkParam("go"))
		{
			h->sendRemotePrivate(p, "!go <arena> (change bot arena)");
		}

	case OP_Moderator:	/* FALL THRU */
		if (c->checkParam("help"))
		{
			h->sendRemotePrivate(p, "!help [topic] (display available commands and what they do)");
			h->sendRemotePrivate(p, "<> required  [] optional  () comments  -s[=assign] switches");
		}
		else if (c->checkParam("listop"))
		{
			h->sendRemotePrivate(p, "!listop (display online operators)");
		}
		else if (c->checkParam("chat"))
		{
			h->sendRemotePrivate(p, "!chat <channel> (set bot chat channel)");
		}
		else if (c->checkParam("version"))
		{
			h->sendRemotePrivate(p, "!version (retrieve version string/maker)");
		}

	case OP_Limited:	/* FALL THRU */
	case OP_Player:
		if (c->checkParam("disabled"))
		{
			h->sendRemotePrivate(p, "Forget about it! This command is disabled");
		}
		break;
	};

	h->imports->talk(makeRemoteHelp(p, c, l));

	c->final = tfinal;
}

void gotRemote(Host *h, char *p, char *m)
{

#ifdef DISABLE_COMMANDS
	return;											// Compiler directive to ignore commands
#endif

	// Retrieve operator level
	opEntry *op = h->botInfo.db->findOperator(p);

	Operator_Level l;

	if (op)
	{
		l = op->getAccess();
	}
	else
	{
		l = OP_Player;
	}

	// Operators with passwords may not use remote private messages [optional]
	if (!h->botInfo.db->remoteOperator)
	{
		if (op->validatePass(""))
		{
			l = OP_Player;
		}
	}

	// Heavens no!
	if (!h->billerOnline) l = OP_Player;

	// Limiter is in effect
	if (h->lowestLevel > OP_Player)
	{
		if (l < h->lowestLevel)
		{
			return;
		}
	}

	// Process command
	Command c(m);									// Wrap in a friendly class

	char *tcmd = c.cmd;							// Backup original pointer in case this is an alias
	h->botInfo.db->aliasCommand(c.cmd);				// Check aliases

try {												// Catch all command-crash bugs

	switch (l)
	{
	case OP_Duke:
	case OP_Baron:
	case OP_King:
	case OP_Emperor:
	case OP_RockStar:
	case OP_Q:
	case OP_God:
	case OP_Owner:
	case OP_SysOp:
		if (c.check("close"))
		{
			h->disconnect(true);
		}

	case OP_SuperModerator:	/* FALL THRU */
		if (c.check("go"))
		{
			if (invalidArena(c.final))
			{
				String s;
				s += "Denied! Invalid arena name \'";
				s += c.final;
				s += "\'";
				h->sendRemotePrivate(p, s);

				break;
			}

			String s;
			s += "Moving to \'";
			s += c.final;
			s += "\'";
			h->sendRemotePrivate(p, s);

			h->changeArena(c.final);
		}

	case OP_Moderator:	/* FALL THRU */
		if (c.check("help"))
		{
			gotRemoteHelp(h, p, &c, l);
		}
		else if (c.check("listop"))
		{
			Operator_Level access = OP_Player;
			Uint32 total = 0;
			String l;

			_listnode <Player> *parse = h->playerlist.head;

			while (parse)
			{
				Player *o = parse->item;
				parse = parse->next;

				if (o->access == OP_Player) continue;

				if (access != o->access)
				{
					l += getLevelString(access = o->access);
					l += ": ";
				}

				l += o->name;

				++total;

				if (parse)
				{
					if (l.len > 80)
					{
						h->sendRemotePrivate(p, l);
						l.clear();
						access = OP_Player;
					}
					else
					{
						l += "  ";
					}
				}
			}

			if (total)
				h->sendRemotePrivate(p, l);
			else
				h->sendRemotePrivate(p, "No operators");
		}
		else if (c.check("chat"))
		{
			if (*c.final)
			{
				if (h->broadcastingErrors)
				{
					h->sendRemotePrivate(p, "Cannot change chat channel while broadcasting errors");
				}
				else
				{
					String s = "?chat=";
					s += c.final;
					h->sendPublic(s);

					s = "chat=";
					s += c.final;
					h->sendRemotePrivate(p, s);
				}
			}
			else
			{
				h->sendRemotePrivate(p, "You must provide channel name(s): !chat squad_chat,newbie_chat,zone_chat");
			}
		}
		else if (c.check("version"))
		{
			h->sendRemotePrivate(p, VERSION_STRING);
		}

	case OP_Limited:	/* FALL THRU */
	case OP_Player:
		if (c.check("lag"))	// unlisted, display session latency
		{
			String s;
			s += "PING Current:";
			s += h->syncPing;
			s += " ms  Average:";
			s += h->avgPing;
			s += " ms  Low:";
			s += h->lowPing;
			s += " ms  High:";
			s += h->highPing;
			s += " ms  Delta:";
			s += h->timeDiff;
			s += " ms";

			h->sendRemotePrivate(p, s);
		}
		else if (c.check("disabled"))
		{
			h->sendRemotePrivate(p, "Forget about it! This command is disabled");
		}
	};

	h->imports->talk(makeRemoteCommand(p, &c, l));

} catch (...)	// Catch command-crash bugs
{
	h->sendRemotePrivate(p, "Oh no!  You've managed to find a command-crash bug!  Report to Catid@pacbell.net");
}

	c.cmd = tcmd;	// Restore pointer from alias string
}
